# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.coherent.innova
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements the drivers for Innova 300 Series gas lasers.


    Implementation Notes
    --------------------

    There are currently 3 drivers implemented Innova300C, ArgonInnova300C
    and KryptonInnova300C. The last two only add to the first the
    corresponding wavelength selection.

    Sources::

        - Innova 300C Manual

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.

"""


from lantz import Q_, Action, Feat, DictFeat
from lantz.errors import InvalidCommand
from lantz.drivers.legacy.serial import SerialDriver


def make_feat(command, **kwargs):

    def get(self):
        return self.query('PRINT {}'.format(command))

    def set(self, value):
        return self.query('{}={}'.format(command, value))

    if kwargs.pop('readonly', None):
        return Feat(fget=get, **kwargs)
    elif kwargs.pop('writeonly', None):
        return Feat(fset=set, **kwargs)

    return Feat(get, set, **kwargs)


class Innova300C(SerialDriver):
    """Innova300 C Series.
    """

    ENCODING = 'ascii'

    SEND_TERMINATION = '\r\n'
    RECV_TERMINATION = '\r\n'


    def __init__(self, port=1, baudrate=1200, **kwargs):
        super().__init__(port, baudrate, bytesize=8, parity='None',
                         stopbits=1, **kwargs)

    def initialize(self):
        super().initialize()
        self.echo_enabled = False

    def query(self, command, *, send_args=(None, None), recv_args=(None, None)):
        """Send query to the laser and return the answer, after handling
        possible errors.

        :param command: command to be sent to the instrument
        :type command: string

        :param send_args: (termination, encoding) to override class defaults
        :param recv_args: (termination, encoding) to override class defaults
        """
        ans = super().query(command, send_args=send_args, recv_args=recv_args)
        # TODO: Echo handling
        if ans == 'Out of Range':
            raise ValueError()
        elif ans.startswith('Syntax Error'):
            raise InvalidCommand()
        elif ans == 'Laser must be off':
            raise Exception('Laser must be off')

        return ans


    # General information and communication

    idn = make_feat('ID',
                    readonly=True,
                    doc='Laser identification, should be I300.',
                    read_once=True)

    software_rev = make_feat('SOFTWARE',
                             readonly=True,
                             doc='Software revision level in the power supply.',
                             read_once=True)

    head_software_rev = make_feat('HEAD SOFTWARE',
                                  readonly=True,
                                  doc='Software revision level in the laser head board.',
                                  read_once=True)

    echo_enabled = make_feat('ECHO',
                             writeonly=True,
                             doc='Echo mode of the serial interface.',
                             values={True: 1, False: 0})

    baudrate = Feat(values={110, 300, 1200, 2400, 4800, 9600, 19200})

    @baudrate.setter
    def baudrate(self, value):
        """RS-232/422 baud rate, the serial connection will be reset after.
        """
        self.query('BAUDRATE={}'.format(value))
        #TODO: RESET Connection


    # Interface

    analog_relative = make_feat('ANALOG MODE',
                                doc='Analog Interface input mode.',
                                values={True: 1, False: 0})

    analog_enabled = make_feat('ANALOGINT',
                               doc='Analog Interface input state.',
                               values={True: 1, False: 0})

    current_range = make_feat('CURRENT RANGE',
                              doc='Current corresponding to 5 Volts at the input'\
                                  ' or output lines of the Analog Interface.',
                              units='A',
                              limits=(10, 100, 1))

    control_pin_high = make_feat('CONTROL',
                                 readonly=True,
                                 doc='State of the input pin 10 of the Analog Interface.',
                                 values={True: 1, False: 0})

    output_pin_high = make_feat('STATUS',
                                doc='State of the output pin 24 and 25 of the Analog Interface.',
                                values={(False, False): 0, (True, False): 1,
                                     (False, True): 2, (True, True): 3})

    # Diagnostics

    @Feat()
    def faults(self):
        """List of all active faults.
        """
        return self.query('PRINT FAULT').split('&')

    autofill_delta = make_feat('AUTOFILL DELTA',
                               readonly=True,
                               doc='Tube voltage minus the autofill setting.',
                               units='V')

    autofill_needed = make_feat('AUTOFILL STATUS',
                                readonly=True,
                                doc='Is the autofill needed (wheter fill is enabled or not)',
                                values={True: 1, False: 0})

    remaining_time = make_feat('HRSTILSHUTDOWN',
                               readonly=True,
                               doc='Number of hours remaining before the laser '\
                                   'will shut down automatically.',
                               units='hour')

    cathode_current = make_feat('CATHODE CURRENT',
                                readonly=True,
                                doc='Laser cathode current (AC).',
                                units='A')

    cathode_voltage = make_feat('CATHODE VOLTAGE',
                                readonly=True,
                                doc='Laser cathode voltage (AC).',
                                units='V')

    time_to_start = make_feat('START',
                              readonly=True,
                              doc='Timer countdown during the start delay cycle.',
                              units='second')

    @Feat()
    def is_in_start_delay(self):
        """Laser is in start delay (tube not ionized)
        """
        return self.query('LASER') == '1'


    tube_time = make_feat('HOURS',
                          readonly=True,
                          doc='Number of operating hours on the plasma tube.',
                          units='hour')

    tube_voltage = make_feat('TUBE VOLTAGE',
                             readonly=True,
                             doc='Laser tube voltage.',
                             units='V')

    water_flow = make_feat('FLOW',
                           readonly=True,
                           doc='Water flow.',
                           units='gallons/minute')

    water_resistivity = make_feat('WATER RESISTIVITY',
                                  readonly=True,
                                  doc='Resistivity of the incoming water to the power supply.',
                                  units='kohm*cm')

    water_temperature = make_feat('WATER TEMPERATURE',
                                  doc='Temperature of the incoming water to the power supply.')


    # Other

    autofill_mode = make_feat('AUTOFILL',
                              doc='Autofill mode.',
                              values={'disabled': 0, 'enabled': 1,
                                   'enabled until next autofill': 2})

    laser_enabled = make_feat('LASER',
                              doc='Energize the power supply.',
                              values={True: 2, False: 0})

    magnet_current = make_feat('MAGNET CURRENT',
                               readonly=True,
                               doc='Laser magnet current.',
                               units='A')

    operating_mode = make_feat('MODE',
                               readonly=True,
                               doc='Laser operating mode.',
                               values={'current regulation': 0,
                                    'reduced bandwidth light regulation': 1,
                                    'standard light regulation': 2,
                                    'current regulation, light regulation out of range': 3})

    # Etalon

    etalon_mode = make_feat('EMODE',
                            doc='Etalon mode.',
                            values={'manual': 0, 'modetrack': 1, 'modetune': 2})

    etalon_temperature = make_feat('ETALON',
                                   readonly=True,
                                   doc='Etalon temperature.',
                                   units='degC')

    @Feat(units='degC', limits=(51.5, 54, 0.001))
    def etalon_temperature_setpoint(self):
        """Setpoint for the etalon temperature.
        """
        return self.query('PRINT SET ETALON')

    @etalon_temperature_setpoint.setter
    def etalon_temperature_setpoint(self, value):
        self.query('ETALON={}'.format(value))


    # Magnetic field

    magnetic_field_high = make_feat('FIELD',
                                    doc='Magnetic field.',
                                    values={True: 1, False: 0})

    @Feat(values={True: 1, False: 0})
    def magnetic_field_setpoint_high(self):
        """Setpoint for magnetic field setting.
        """
        return self.query('PRINT SET FIELD')

    @magnetic_field_setpoint_high.setter
    def magnetic_field_setpoint_high(self, value):
        self.query('FIELD={}'.format(value))


    # Light and current regulation

    powertrack_mode_enabled = make_feat('PT',
                                        doc='PowerTrack.',
                                        values={True: 1, False: 0})

    @DictFeat(keys=('A', 'B'), limits=(0, 255))
    def powertrack_position(self, key):
        """Relative position of the PowerTrack solenoids.
        """
        return self.query('PRINT PTDAC{}'.format(key))

    @powertrack_position.setter
    def powertrack_position(self, key, value):
        self.query('PTDAC{}={}'.format(key, value))

    @Action()
    def recalibrate_powertrack(self):
        """Recalibrate PowerTrack. This will only execute if PowerTrack is on
        and light regulation is off
        """
        self.query('PT=2')

    @Action()
    def center_powertrack(self):
        """Center PowerTrack and turn it off.
        """
        self.query('PT=3')

    current = make_feat('CURRENT',
                        readonly=True,
                        doc='Current regulation mode.',
                        units='A')

    @Feat(units='A', limits=(0, 50, 0.01))
    def current_setpoint(self):
        """Current setpoint when using the current regulation mode.
        """
        return self.query('PRINT SET CURRENT')

    @current_setpoint.setter
    def current_setpoint(self, value):
        self.query('CURRENT={}'.format(value))


    power = make_feat('LIGHT 3',
                      readonly=True,
                      doc='Current power output.',
                      units='A')

    @Feat(units='W', limits=(0, 50, 0.0001))
    def power_setpoint(self):
        """Setpoint for the light regulation.
        """
        return self.query('PRINT SET LIGHT')

    @power_setpoint.setter
    def power_setpoint(self, value):
        self.query('LIGHT={}'.format(value))


    auto_light_cal_enabled = make_feat('AUTOLTCAL',
                                       doc='Automatic light regulation calibration flag.',
                                       values={True: 1, False: 0})

    current_change_limit = make_feat('PCTCHGTILRECAL',
                                     doc='Percent tube change before an automatic '\
                                         'light regulation recalibration becomes '\
                                         'necessary.',
                                     units='', #TODO: %
                                     limits=(5, 100, 1))


class ArgonInnova300C(Innova300C):
    """Argon Innova 300C.
    """

    wavelength = make_feat('WAVELENGTH',
                           doc='Wavelength for the internal power meter calibration',
                           values={351, 364, 454, 457, 465, 472, 476, 488, 496, 501,
                                 514, 528, 1090, 'MLVS', 'MLUV', 'MLDUV'})


class KryptonInnova300C(Innova300C):
    """Krypton Innova 300C.
    """
    wavelength = make_feat('WAVELENGTH',
                           doc='Wavelength for the internal power meter calibration',
                           values={476, 482, 520, 530, 568, 647, 676, 752, 'MLVS',
                                 'MLUV', 'MLVI', 'MLBG', 'MLRD', 'MLIR'})

if __name__ == '__main__':
    import argparse
    import lantz.log

    parser = argparse.ArgumentParser(description='Test Kentech HRI')
    parser.add_argument('-i', '--interactive', action='store_true',
                        default=False, help='Show interactive GUI')
    parser.add_argument('-p', '--port', type=str, default='17',
                        help='Serial port to connect to')

    args = parser.parse_args()
    lantz.log.log_to_socket(lantz.log.DEBUG)
    with Innova300C(args.port) as inst:
        if args.interactive:
            from lantz.ui.app import start_test_app
            start_test_app(inst)
        else:
            print(inst.idn)
            print(inst.software_rev)
            print(inst.head_software_rev)



# -*- coding: utf-8 -*-
"""
    lantz.drivers.examples.fungen
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements the Signal Generator described in the Lantz tutorial.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from lantz import Feat, DictFeat, Action
from lantz import errors
from lantz import MessageBasedDriver


class LantzSignalGenerator(MessageBasedDriver):
    """Lantz Signal Generator
    """

    DEFAULTS = {'COMMON': {'write_termination': '\n',
                           'read_termination': '\n'}}


    def query(self, command, *, send_args=(None, None), recv_args=(None, None)):
        answer = super().query(command, send_args=send_args, recv_args=recv_args)
        if answer == 'ERROR':
            raise errors.InstrumentError
        return answer

    @Feat(read_once=True)
    def idn(self):
        return self.query('?IDN')

    @Feat(units='V', limits=(10,))
    def amplitude(self):
        """Amplitude.
        """
        return float(self.query('?AMP'))

    @amplitude.setter
    def amplitude(self, value):
        self.query('!AMP {:.1f}'.format(value))

    @Feat(units='V', limits=(-5, 5, .01))
    def offset(self):
        """Offset.
        """
        return float(self.query('?OFF'))

    @offset.setter
    def offset(self, value):
        self.query('!OFF {:.1f}'.format(value))

    @Feat(units='Hz', limits=(1, 1e+5))
    def frequency(self):
        """Frequency.
        """
        return float(self.query('?FRE'))

    @frequency.setter
    def frequency(self, value):
        self.query('!FRE {:.2f}'.format(value))

    @Feat(values={True: 1, False: 0})
    def output_enabled(self):
        """Analog output enabled.
        """
        return int(self.query('?OUT'))

    @output_enabled.setter
    def output_enabled(self, value):
        self.query('!OUT {}'.format(value))

    @Feat(values={'sine': '0', 'square': '1', 'triangular': '2', 'ramp': '3'})
    def waveform(self):
        return self.query('?WVF')

    @waveform.setter
    def waveform(self, value):
        self.query('!WVF {}'.format(value))

    @DictFeat(values={True: '1', False: '0'}, keys=list(range(1,9)))
    def dout(self, key):
        """Digital output state.
        """
        return self.query('?DOU {}'.format(key))

    @dout.setter
    def dout(self, key, value):
        self.query('!DOU {} {}'.format(key, value))

    @DictFeat(values={True: '1', False: '0'}, keys=list(range(1,9)))
    def din(self, key):
        """Digital input state.
        """
        return self.query('?DIN {}'.format(key))

    @Action()
    def calibrate(self):
        """Calibrate.
        """
        self.query('!CAL')

    @Action()
    def self_test(self, level=1, repetitions=3):
        """Reset to .
        """
        self.query('!TES {} {}'.format(level, repetitions))



from lantz import Feat, DictFeat, Action
from lantz.driver import Driver
from collections import OrderedDict
import time
import ripy
import numpy as np

class RI(Driver):

    gain_values = OrderedDict([
        ('High', True),
        ('Low', False),
    ])

    antialias_values = OrderedDict([
        ('On', True),
        ('Off', False),
    ])

    def __init__(self):
        super(RI, self).__init__()
        self._ripy = ripy.Device()
        self._ripy.open()

    @Feat(units = "Hz")
    def sample_rate(self):
        return self._ripy.samplerate

    @Feat()
    def usb_speed(self):
        return self._ripy.usb_speed

    @Feat(values = gain_values)
    def gain(self):
        """
        Get the current gain setting
        """
        return self._ripy.highgain

    @gain.setter
    def gain(self, gain_value):
        """
        Set the current gain setting \n
        It can be \"High\" or \"Low\"
        """
        self._ripy.set_highgain(gain_value)

    @Feat(values = antialias_values)
    def antialias(self):
        """
        Get the current antialias setting
        """
        return self._ripy.antialias


    @antialias.setter
    def antialias(self, antialias_value):
        """
        Set the current gain setting \n
        It can be \"On\" or \"Off\"
        """
        self._ripy.set_antialias(antialias_value)

    @Action()
    def get_raw_data(self, nsamples):
        """
        Get n many samples
        """
        return self._ripy.get_raw_data(nsamples)


    @Action()
    def get_triggered_data(self, nsamples, ntimes=1, key="T"):
        """
        Get n many samples ntimes many trigger detects a rising edge
        """
        total = nsamples*ntimes
        raw_return = self._ripy.get_raw_data(total, trig_port = key, trig_mode = "rising", samples_per_trigger = nsamples)
        if ntimes == 1:
            return raw_return
        else:
            return np.hsplit(raw_return,ntimes)




from lantz import Feat, DictFeat, Action
from lantz.messagebased import MessageBasedDriver

class BK9129b(MessageBasedDriver):

    DEFAULTS = {
        'ASRL': {
            'write_termination': '\n',
            'read_termination': '\n',
            'timeout': 1000,
        }
    }

    @Feat()
    def idn(self):
        return self.query('*IDN?')

    @Action()
    def remote(self):
        self.write('SYST:REM')

    @Action()
    def local(self):
        self.write('SYST:LOC')

    @Feat()
    def channel(self):
        return int(self.query('INST?').lstrip('CH'))

    @channel.setter
    def channel(self, value):
        self.write('INST CH{}'.format(value))

    @DictFeat(units='A')
    def current(self, ch):
        return self.query('MEAS:CURR?')

    @DictFeat(units='V')
    def voltage(self, ch):
        return self.query('MEAS?')

    @voltage.setter
    def voltage(self, value):
        self.write('')

    @Feat(units='V')
    def voltage_limit(self):
        return self.query('VOLT:LIMIT?')

    @voltage_limit.setter
    def voltage_limit(self, value):
        self.write('VOLT:LIMIT {:1.2f}'.format(value))

    @Feat(values={False: 0, True: 1})
    def state(self):
        return int(self.query('CHAN:OUTP?'))

    @state.setter
    def state(self, value):
        self.write('CHAN:OUTP {:d}'.format(value))

def main():
    import logging
    import sys
    from lantz.log import log_to_screen
    import numpy as np
    log_to_screen(logging.CRITICAL)
    res_name = sys.argv[1]
    with BK9129b(res_name) as inst:
        print(inst.idn)
        inst.remote()
        inst.channel = 2
        print(inst.channel)
        print(inst.current[1])
        print(inst.voltage[1])
        inst.state = True
        print(inst.state)
        inst.voltage_limit = 1.23
        print(inst.voltage_limit)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
    lantz.drivers.tektronix.afg3021b
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements the drivers to control a signal generator.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from lantz import Feat

from lantz import MessageBasedDriver



class AFG3021b(MessageBasedDriver):

    MANUFACTURER_ID = '0x0699'
    MODEL_CODE = '0x0346'

    @Feat()
    def idn(self):
        return inst.query('*IDN?')


if __name__ == '__main__':
    import argparse
    import lantz.log

    parser = argparse.ArgumentParser(description='Test Kentech HRI')
    parser.add_argument('-i', '--interactive', action='store_true',
                        default=False, help='Show interactive GUI')
    parser.add_argument('-p', '--port', type=str, default='17',
                        help='Serial port to connect to')

    args = parser.parse_args()
    lantz.log.log_to_socket(lantz.log.DEBUG)

    with AFG3021b('USB0::0x0699::0x0346::C033250::INSTR') as inst:
        print(inst.idn)

    with AFG3021b.from_hostname('192.168.0.1') as inst:
        print(inst.idn)

    with AFG3021b.from_usbtmc(serial_number='C033250') as inst:
        print(inst.idn)



# -*- coding: utf-8 -*-
"""
    lantz.drivers.alliedvision
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implementation for a alliedvision camera via pymba and pylon

    Requires:
    - pymba: https://github.com/morefigs/pymba
    - vimba: https://www.alliedvision.com/en/products/software.html

    Log:
    - create same API as lantz.drivers.basler


    Author: Vasco Tenner
    Date: 20181204

    TODO:
    - Test
    - 12 bit packet readout
    - Bandwith control
    - Dynamically set available values for feats
"""

from lantz.driver import Driver
from lantz import Feat, DictFeat, Action
from pymba import Vimba
import numpy as np
import threading
import time


beginner_controls = ['ExposureTimeAbs', 'GainRaw', 'Width', 'Height',
                     'OffsetX', 'OffsetY']
aliases = {'exposure_time': 'ExposureTimeAbs',
           'gain': 'GainRaw',
          }


def todict(listitems):
    'Helper function to create dicts usable in feats'
    d = {}
    for item in listitems:
        d.update({item: item})
    return d


def attach_dyn_propr(instance, prop_name, propr):
    """Attach property proper to instance with name prop_name.

    Reference:
    * https://stackoverflow.com/a/1355444/509706
    * https://stackoverflow.com/questions/48448074
    """
    class_name = instance.__class__.__name__ + 'C'
    child_class = type(class_name, (instance.__class__,), {prop_name: propr})

    instance.__class__ = child_class


def create_getter(p):
    def tmpfunc(self):
        return self.cam[p]
    return tmpfunc


def create_setter(p):
    def tmpfunc(self, val):
        self.cam[p] = val
    return tmpfunc


def list_cameras():
    with Vimba() as vimba:
        # get system object
        system = vimba.getSystem()

        # list available cameras (after enabling discovery for GigE cameras)
        if system.GeVTLIsPresent:
            system.runFeatureCommand("GeVDiscoveryAllOnce")
            time.sleep(0.2)
        cameraIds = vimba.getCameraIds()
        for cameraId in cameraIds:
            print(cameraId)


class VimbaCam(Driver):

    def __init__(self, camera=0, level='beginner',
                 *args, **kwargs):
        """
        @params
        :type camera_num: int, The camera device index: 0,1,..
        :type level: str, Level of controls to show ['beginner', 'expert']

        Example:
        import lantz
        from lantz.drivers.alliedvision import Cam
        import time
        try:
                lantzlog
        except NameError:
                lantzlog = lantz.log.log_to_screen(level=lantz.log.DEBUG)

        cam = Cam(camera='Basler acA4112-8gm (40006341)')
        cam.initialize()
        cam.exposure_time
        cam.exposure_time = 3010
        cam.exposure_time
        cam.grab_image()
        print('Speedtest:')
        nr = 10
        start = time.time()
        for n in range(nr):
            cam.grab_image()
            duration = (time.time()-start)*1000*lantz.Q_('ms')
            print('Read {} images in {}. Reading alone took {}. Framerate {}'.format(nr,
                duration, duration - nr* cam.exposure_time, nr / duration.to('s')))
        cam.finalize()
        """
        super().__init__(*args, **kwargs)
        self.camera = camera
        self.level = level
        # Some actions cannot be performed while reading
        self._grabbing_lock = threading.RLock()

    def initialize(self):
        """
        Params:
        camera -- number in list of show_cameras or friendly_name
        """

        self.vimba = Vimba()
        self.vimba.startup()

        cameras = self.vimba.getCameraIds()
        self.log_debug('Available cameras are:' + str(cameras))

        try:
            if isinstance(self.camera, int):
                cam = cameras[self.camera]
                self.cam = self.vimba.getCamera(cam)
            else:
                self.cam = self.vimba.getCamera(self.camera)
        except RuntimeError as err:
            self.log_error(err)
            raise RuntimeError(err)

        self.frame = self.cam.getFrame()
        self.frame.announceFrame()
        self.cam.startCapture()

        self._dynamically_add_properties()
        self._aliases()

        # get rid of Mono12Packed and give a log error:
        fmt = self.pixel_format
        if fmt == str('Mono12Packed'):
            self.log_error('PixelFormat {} not supported. Using Mono12 '
                           'instead'.format(fmt))
            self.pixel_format = 'Mono12'

        # Go to full available speed
        # cam.properties['DeviceLinkThroughputLimitMode'] = 'Off'

    def finalize(self):
        self.cam.endCapture()
        self.cam.revokeAllFrames()
        return

    def _dynamically_add_properties(self):
        """Add all properties available on driver as Feats"""
        props = self.getFeatureNames() if self.level == 'expert' else beginner_controls
        for p in props:
            info = self.cam.getFeatureInfo(p)
            range_ = self.cam.getFeatureRange(p)
            limits = range_ if isinstance(tuple, range_) else None
            values = range_ if isinstance(list, range_) else None

            feat = Feat(fget=create_getter(p),
                        fset=create_setter(p),
                        doc=info.description,
                        units=info.unit,
                        limits=limits,
                        values=values,
                        )
            feat.name = p
            attach_dyn_propr(self, p, feat)

    def _aliases(self):
        """Add easy to use aliases to strange internal pylon names

        Note that in the Logs, the original is renamed to the alias"""
        for alias, orig in aliases.items():
            attach_dyn_propr(self, alias, self.feats[orig].feat)

    @Feat()
    def info(self):
        # We can still get information of the camera back
        return 'Camera info of camera object:', self.cam.getInfo()  # TODO TEST

    # Most properties are added automatically by _dynamically_add_properties

    @Feat(values=todict(['Mono8', 'Mono12', 'Mono12Packed']))
    def pixel_format(self):
        fmt = self.cam['PixelFormat']
        if fmt == 'Mono12Packed':
            self.log_error('PixelFormat {} not supported. Use Mono12 instead'
                           ''.format(fmt))
        return fmt

    @pixel_format.setter
    def pixel_format(self, value):
        if value == 'Mono12Packed':
            self.log_error('PixelFormat {} not supported. Using Mono12 '
                           'instead'.format(value))
            value = 'Mono12'
        self.cam['PixelFormat'] = value

    @Feat()
    def properties(self):
        """Dict with all properties supported by pylon dll driver"""
        return self.cam.getFeatureNames()

    @Action()
    def list_properties(self):
        """List all properties and their values"""
        for key in self.cam.getFeatureNames():
            try:
                value = self.cam[key]
            except IOError:
                value = '<NOT READABLE>'

            description = self.cam.getFeatureInfo(key)
            range_ = self.cam.getFeatureRange(key)
            print('{0} ({1}):\t{2}\t{3}'.format(key, description, value, range_))

    @Action(log_output=False)
    def grab_image(self):
        """Record a single image from the camera"""
        self.cam.AcquisitionMode = 'SingleFrame'
        try:
            self.frame.queueFrameCapture()
            success = True
        except:
            success = False

        self.cam.runFeatureCommand('AcquisitionStart')
        self.cam.runFeatureCommand('AcquisitionStop')
        self.frame.waitFrameCapture(0)
        frame_data = self.frame.getBufferByteData()
        if success:
            img_config = {
                'buffer': frame_data,
                'dtype': np.uint8,
                'shape': (self.frame.height, self.frame.width, 1),
            }
            img = np.ndarray(**img_config)
            return img[..., 0]
        else:
            return None

    @Action(log_output=False)
    def grab_images(self, num=1):
        # ΤΟDO see https://gist.github.com/npyoung/1c160c9eee91fd44c587
        raise NotImplemented()
        with self._grabbing_lock:
            img = self.cam.grab_images(num)
        return img

    @Action(log_output=False)
    def getFrame(self):
        """Backward compatibility"""
        return self.grab_image

    @Action()
    def set_roi(self, height, width, yoffset, xoffset):
        # Validation:
        if width + xoffset > self.properties['WidthMax']:
            self.log_error('Not setting ROI:  Width + xoffset = {} exceeding '
                           'max width of camera {}.'.format(width + xoffset,
                                                            self.properties['WidthMax']))
            return
        if height + yoffset > self.properties['HeightMax']:
            self.log_error('Not setting ROI: Height + yoffset = {} exceeding '
                           'max height of camera {}.'.format(height + yoffset,
                                                             self.properties['HeightMax']))
            return

        # Offset should be multiple of 2:
        xoffset -= xoffset % 2
        yoffset -= yoffset % 2

        if height < 16:
            self.log_error('Height {} too small, smaller than 16. Adjusting '
                           'to 16'.format(height))
            height = 16
        if width < 16:
            self.log_error('Width {} too small, smaller than 16. Adjusting '
                           'to 16'.format(width))
            width = 16

        with self._grabbing_lock:
            # Order matters!
            if self.OffsetY > yoffset:
                self.Height = height
                self.OffsetY = yoffset
            else:
                self.Height = height
                self.OffsetY = yoffset
            if self.OffsetX > xoffset:
                self.OffsetX = xoffset
                self.Width = width
            else:
                self.Width = width
                self.OffsetX = xoffset

    @Action()
    def reset_roi(self):
        """Sets ROI to maximum camera size"""
        self.set_roi(self.properties['HeightMax'],
                     self.properties['WidthMax'],
                     0,
                     0)

    # Helperfunctions for ROI settings
    def limit_width(self, dx):
        if dx > self.properties['WidthMax']:
            dx = self.properties['WidthMax']
        elif dx < 16:
            dx = 16
        return dx

    def limit_height(self, dy):
        if dy > self.properties['HeightMax']:
            dy = self.properties['HeightMax']
        elif dy < 16:
            dy = 16
        return dy

    def limit_xoffset(self, xoffset, dx):
        if xoffset < 0:
            xoffset = 0
        if xoffset + dx > self.properties['WidthMax']:
            xoffset = self.properties['WidthMax'] - dx
        return xoffset

    def limit_yoffset(self, yoffset, dy):
        if yoffset < 0:
            yoffset = 0
        if yoffset + dy > self.properties['HeightMax']:
            yoffset = self.properties['HeightMax'] - dy
        return yoffset

    @Action()
    def calc_roi(self, center=None, size=None, coords=None):
        """Calculate the left bottom corner and the width and height
        of a box with center (x,y) and size x [(x,y)]. Respects device
        size"""
        if center and size:
            y, x = center
            try:
                dy, dx = size
            except TypeError:
                dx = dy = size

            # Make sizes never exceed camera sizes
            dx = self.limit_width(dx)
            dy = self.limit_width(dy)

            xoffset = x - dx // 2
            yoffset = y - dy // 2

            xoffset = self.limit_xoffset(xoffset, dx)
            yoffset = self.limit_yoffset(yoffset, dy)

            return dy, dx, yoffset, xoffset

        elif coords:
            xoffset = int(coords[1][0])
            dx = int(coords[1][1] - xoffset)

            yoffset = int(coords[0][0])
            dy = int(coords[0][1] - yoffset)

            # print(dy,dx)
            dx = self.limit_width(dx)
            dy = self.limit_height(dy)

            # print(yoffset, xoffset)
            xoffset = self.limit_xoffset(xoffset, dx)
            yoffset = self.limit_yoffset(yoffset, dy)

            return dy, dx, yoffset, xoffset

        else:
            raise ValueError('center&size or coords should be supplied')

    def calc_roi_from_rel_coords(self, relcoords):
        """Calculate the new ROI from coordinates relative to the current
        viewport"""

        coords = ((self.OffsetY + relcoords[0][0],
                   self.OffsetY + relcoords[0][1]),
                  (self.OffsetX + relcoords[1][0],
                   self.OffsetX + relcoords[1][1]))
        # print('Rel_coords says new coords are', coords)
        return self.calc_roi(coords=coords)


from PyQt5 import QtCore
import ctypes

from lantz.driver import Driver

xinput = ctypes.windll.xinput9_1_0

class XINPUT_GAMEPAD(ctypes.Structure):
    _fields_ = [
        ('buttons', ctypes.c_ushort),
        ('left_trigger', ctypes.c_ubyte),
        ('right_trigger', ctypes.c_ubyte),
        ('l_thumb_x', ctypes.c_short),
        ('l_thumb_y', ctypes.c_short),
        ('r_thumb_x', ctypes.c_short),
        ('r_thumb_y', ctypes.c_short),
    ]

class XINPUT_STATE(ctypes.Structure):
    _fields_ = [
        ('packet_number', ctypes.c_ulong),
        ('gamepad', XINPUT_GAMEPAD),
    ]

class XINPUT_VIBRATION(ctypes.Structure):
    _fields_ = [
        ('wLeftMotorSpeed', ctypes.c_ushort),
        ('wRightMotorSpeed', ctypes.c_ushort),
    ]

def val_to_bits(value, nbits=16):
    return [int(v) for v in bin(value)[2:].zfill(nbits)]

ERROR_DEVICE_NOT_CONNECTED = 1167
ERROR_SUCCESS = 0

class _XInputController(QtCore.QObject):

    max_devices = 4

    on_axis = QtCore.pyqtSignal(str, float)
    on_button = QtCore.pyqtSignal(int, bool)

    def __init__(self, device_number, input_deadzone=0.01):
        self.device_number = device_number
        self.input_deadzone = input_deadzone
        super().__init__()
        self._last_state = self.get_state()
        return

    @staticmethod
    def enumerate_devices():
        devices = (XInputController(device_id)
                   for device_id in range(XInputController.max_devices))
        return [d for d in devices if d.is_connected()]

    def normalize(self, value, sizeof_ctype):
        bitsof_ctype = 8 * sizeof_ctype
        return value / (2 ** bitsof_ctype - 1)

    def is_connected(self):
        return self._last_state is not None

    def set_vibration(self, left_motor, right_motor):
        XInputSetState = xinput.XInputSetState
        XInputSetState.argtypes = [ctypes.c_uint, ctypes.POINTER(XINPUT_VIBRATION)]
        XInputSetState.restype = ctypes.c_uint

        left_value = int(left_motor * 65535)
        right_value = int(right_motor * 65535)
        vibration = XINPUT_VIBRATION(left_value, right_value)
        XInputSetState(self.device_number, ctypes.byref(vibration))
        return

    def get_state(self):
        state = XINPUT_STATE()
        result = xinput.XInputGetState(self.device_number, ctypes.byref(state))
        if result == ERROR_SUCCESS:
            return state
        elif result == ERROR_DEVICE_NOT_CONNECTED:
            return None
        else:
            raise RuntimeError()

    def dispatch_events(self):
        state = self.get_state()
        if state is None:
            raise RuntimeError()
        if self._last_state is not None and state.packet_number != self._last_state.packet_number:
            self.handle_changed_state(state)
        self._last_state = state
        return
        
    def handle_changed_state(self, state):
        self.dispatch_axis_events(state)
        self.dispatch_button_events(state)
        return

    def dispatch_axis_events(self, state):
        axis_fields = dict(XINPUT_GAMEPAD._fields_)
        axis_fields.pop('buttons')
        for axis, ctype in axis_fields.items():
            old_value = getattr(self._last_state.gamepad, axis)
            new_value = getattr(state.gamepad, axis)
            sizeof_ctype = ctypes.sizeof(ctype)
            old_value = self.normalize(old_value, sizeof_ctype)
            new_value = self.normalize(new_value, sizeof_ctype)
            d_value = abs(new_value - old_value)
            if d_value:
                self.on_axis.emit(axis, new_value)

    def dispatch_button_events(self, state):
        changed = state.gamepad.buttons ^ self._last_state.gamepad.buttons
        changed_bits = val_to_bits(changed)
        state_bits = val_to_bits(state.gamepad.buttons)
        button_numbers = range(1, 17)
        changed_buttons = [(bn, bs) for changed, bn, bs in zip(changed_bits, button_numbers, state_bits)
                           if changed]
        [self.on_button.emit(button, bool(pressed)) for button, pressed in changed_buttons]
        return

    def axis_event(self, axis, value):
        pass

    def button_event(self, button, pressed):
        pass

    def shutdown(self, offset=103):
        shutdown_f_type = ctypes.CFUNCTYPE(ctypes.c_int)
        shutdown_f = shutdown_f_type(xinput._handle + offset)
        shutdown_f(self.device_number)
        return

class XInputController(Driver):

    def __init__(self, device_number, input_deadzone=0.01):
        self.controller = _XInputController(device_number, input_deadzone)
        return

    def set_axis_callback(self, f):
        try:
            self.controller.on_axis.disconnect()
        except TypeError:
            pass
        self.controller.on_axis.connect(f)
        return

    def set_button_callback(self, f):
        try:
            self.controller.on_button.disconnect()
        except TypeError:
            pass
        self.controller.on_button.connect(f)
        return

    def dispatch_events(self):
        self.controller.dispatch_events()
        return
